/*
 * Chromosone Definition File
 * The chromosone is defined as a object, with a pointer to a list
 * of genes (Gene objects, see Gene.h )
 * Methods include Cut and Splice
 *
 * Chromosone    is an instance of a Chromosone
 * PChromosone   is a pointer to an instance of a Chromosone
 *
 * John Dunkelberg
 * created: 19 Feb 1994
 */

#include "gene.H"

extern int DEBUG;

class Chromosone
{
  long chrom_id;    // identification tag of the chromosone

  int genelen;      // length of the gene-string (number of genes)
  int uniqlen;      // number of unique genes
  int mark;         // has this chromosone already been tourney'd

  double fitness;   // fitness of the chromosone

  Gene *first_gene;  // first gene in the chromosone
  Gene *last_gene;   // last gene in the chromosone

  Chromosone *next_chrom;  // next chromosone in line

public:
// Construct and Destruct Functions
  Chromosone(void);     // constructor for the chromosone
  ~Chromosone(void);    // destructor for the chromosone
  int Kill(void);       // kill all Chromosones in our list
  int Assign(long, Gene *);  // assign a gene sequence to the chromosone

// Fitness setting function (friend of problem)
  double Fitness(Problem *, Chromosone *);

// Meta Functions
  int Connect(Chromosone *);    // add another chrom to our list
  Chromosone *Disconnect();    // cut off our tail and pass it back
  Chromosone *Next();          // pass back location of our tail

// Private Info Functions
  int Get_Len(void)
    { return genelen; }  // returns the genelen of the chromosone
  int Get_Uniq(void)
    { return uniqlen; }  // returns the number of unique genes
  double Get_Fit(void)
    { return fitness; }  // returns the fitness of the chromosone
  int Get_Mark(void)
    { return mark; }     // returns the tourney marker on the chromosone
  int Mark(void)
    { mark = 1; return mark; }  // mark the chromosone
  int Unmark(void)
    { mark = 0; return mark; }  // unmark the chromosone
  int Unmark_All(void);         // unmark all the chromosones

// Internal Functions
  Chromosone *Copy(void); // creates a copy of this chromosone
  int Threshold(Chromosone *);  // checks to see that we can threshold
  Chromosone *Cut(long); // returns the new cut half of the chromosone
  void Splice(Chromosone *);   // make spliced chromosone
  void Show(int);     // provide information of variable depth on chrom.
  void Show(void);    // overload of above, with 0 debug_level
};

typedef Chromosone *PChromosone;

// Construct a Chromosone
 Chromosone::Chromosone()
{
  genelen = 0;
  uniqlen = 0;
  fitness = -1.0;
  mark = 0;
  
  first_gene = last_gene = (Gene *) NULL;
  next_chrom = (Chromosone *) NULL;
}

// Destroy a Chromosone
 Chromosone::~Chromosone(void)
{
  if (first_gene != (Gene *) NULL)
    {
      first_gene->Kill();
      delete(first_gene);
    }
}

// Kill all the Chromosones in the list after this one
int Chromosone::Kill(void)
{
  if (next_chrom == (Chromosone *) NULL)
    return(0);

  next_chrom->Kill();
  delete(next_chrom);
  next_chrom = (Chromosone *) NULL;
  return(1);
}

// Assign a gene sequence to a blank chromosone
int Chromosone::Assign(long name_id, Gene *new_gene)
{
  if (first_gene != (Gene *) NULL)
    {
      cout << "Attempt to Assign to a non-blank chromosone, Blocked\n";
      return(0);
    }
  if (new_gene == (Gene *) NULL)
    {
      cout << "Attempt to Assign a NULL gene, Blocked\n";
      return(0);
    }
  
  Gene *cur_gene;
  int gene_count;
  
  // name the chromosone
  chrom_id = name_id;
  
  // attach the new gene sequence
  first_gene = new_gene;

  // pace thru the new gene sequence to find the end
  for(cur_gene = new_gene, gene_count = 1;
      (cur_gene->Next() != (Gene *) NULL);
      cur_gene = cur_gene->Next(), gene_count++);
  
  // assign the last_gene marker
  last_gene = cur_gene;
  // assign the value of the gene length
  genelen = gene_count;
  
  return(gene_count);
}

// Fitness calculation functions
// Assigns internal fitness and returns value
double Chromosone::Fitness(Problem *problem, Chromosone *tplate)
{
  Chromosone *temp_chrom;    // temporary chromosone
  int *map;                 // fixed size mapping
  
  // Make temp_chrom a composite of the tplate and test_chrom
  temp_chrom = tplate->Copy();
  temp_chrom->Splice(this);
  
  // DEBUG
  // cout << "Temp_chrom\n";
  // temp_chrom->Show(DEBUG);

  // make the mapping from the test_chromosone
  int count, maxlen;
  Gene *cur_gene;
  maxlen = temp_chrom->Get_Len();
  if ((map = (int *) malloc(maxlen * sizeof(int))) == (int *) NULL)
    {
      cout << "Chromosone::Fitness memory allocation failed for map\n";
      exit(0);
    }
  for(count=0,cur_gene = temp_chrom->first_gene;
      (count < maxlen);
      count++, cur_gene = cur_gene->Next())
      {
	if (cur_gene == (Gene *) NULL)
	  {
	    cout << "Problem::Fitness got an unexpected NULL gene\n";
	    exit(0);
	  }
	map[cur_gene->Gene_id()] = cur_gene->Allele();
      }

  // DEBUG
  //  for(count=0;(count<maxlen);count++)
  //    cout << "[" << count << "," << map[count] << "]" << endl;

  // Now we have a completed mapping array, so now let's use it
  // to get the fitness

  int node,side,target;
  int x1,x2,y1,y2,dx,dy,temp;
  double maxfit,newfit=0;
  int edges = 0;

  for(node=0;(node < problem->Get_Size());node++)
    for(side=0;(side < problem->Get_Connect());side++)
      {
	// edge[node][side]
	target = problem->Get_Edge(node,side);

	// if it's not a null edge, then compute change in fitness
	if (target != -1)
	  {
	    edges++;    // realize the mapping has an edge

	    // Modified Manhattan Distance
	    x1 = map[node]%(problem->Get_Mach_Side());
	    x2 = map[target]%(problem->Get_Mach_Side());
	    dx = abs(x1 - x2);
	    temp = ((x1>x2) ? x2+abs((problem->Get_Mach_Side())-x1) :
		    x1+abs((problem->Get_Mach_Side())-x2));
	    dx = ((dx>temp) ? temp : dx);

	    y1 = map[node]/(problem->Get_Mach_Side());
	    y2 = map[target]/(problem->Get_Mach_Side());
	    dy = abs(y1 - y2);
	    temp = ((y1>y2) ? y2 + abs((problem->Get_Mach_Side())-y1) :
		    y1 + abs((problem->Get_Mach_Side())-y2));
	    dy = ((dy>temp) ? temp : dy);
	    newfit += ((dx>dy) ? dy : dx) + abs(dx-dy);

	    //cout << "\n ("<<x1<<","<<y1<<") ("<<x2<<","<<y2<<") ";
	    //cout << "dx"<<dx<<" dy"<<dy<<" tot"<<newfit;
	  }
      }
  fitness = newfit/2.0;       // We count every edge twice, so rectify
  edges = edges/2;            // ditto

  // squash our fitness into the range [0,1]
  maxfit = (double) (edges * ((problem->Get_Mach_Side())/2));

  // DEBUG
  //  cout << "maxfit " << maxfit << " fitness " << fitness << endl;

  if (maxfit == ((double) edges))
    fitness = 1.0;
  else
    fitness = 1.0 - (fitness / maxfit);
  return(fitness);
}

// add another chrom to our list
int Chromosone::Connect(Chromosone *new_tail)
{
  if (new_tail == (Chromosone *) NULL)
    return(0);

  next_chrom = new_tail;
  return(1);
}

// cut off our tail and pass it back
Chromosone *Chromosone::Disconnect()
{
  Chromosone *tail;

  tail = next_chrom;
  next_chrom = (Chromosone *) NULL;
  return(tail);
}

// pass back location of our tail
Chromosone *Chromosone::Next()
{
  return(next_chrom);
}

// Unmark all the Chromosones in the list after this one
int Chromosone::Unmark_All(void)
{
  if (next_chrom == (Chromosone *) NULL)
    return(0);
  next_chrom->Unmark_All();
  mark = 0;
  return(1);
}

// Make a copy of the Chromosone
// A copied Chromosone looks just like the original except it is 
// not marked and has a NULL next pointer
Chromosone *Chromosone::Copy(void)
{
  Chromosone *new_chrom;

  // create space for the cloned chromosone
  new_chrom = new Chromosone;

  // Copy over constant values to the clone
  new_chrom->chrom_id = chrom_id;
  new_chrom->genelen = genelen;
  new_chrom->uniqlen = uniqlen;
  new_chrom->fitness = fitness;
  new_chrom->next_chrom = (Chromosone *) NULL;

  // Copys are not MARKed from their original!

  // copy over all the genes to the cloned chromosone
  Gene *cur_gene, *eugene;

  new_chrom->first_gene = first_gene->Copy();
  for(cur_gene=first_gene->Next(), eugene = new_chrom->first_gene;
      (cur_gene != (Gene *) NULL);
      cur_gene=cur_gene->Next())
      {
	eugene->Connect(cur_gene->Copy());
	eugene = eugene->Next();
      }
  new_chrom->last_gene = eugene;

  return(new_chrom);
}

// Threshold checking to make sure two chromosones are similar enough
// to be competing in the tournament
int Chromosone::Threshold(Chromosone *)
{
  return(1);
}

// Cut the Chromosone in two, and return the old tail section
Chromosone *Chromosone::Cut(long new_id)
{
  int cut_point;
  Gene *cur_gene, *tail_gene;
  Chromosone *new_chrom;
  
  new_chrom = new Chromosone;
  if (new_chrom == (Chromosone *) NULL)
    {
      cout << "Failure to Allocate Memory for Chromosone" << new_id << endl;
      exit(1);
    }

  // Choose a cut_point, make sure the target gene will be nonzero length
  cut_point = ((int) random()) % (genelen-1);
  
  // rectify the length values of the parent chromosones
  genelen = cut_point+1;

  // run through this chromosone's genes to the cut_point
  for(cur_gene= first_gene;
      (cut_point);
      cut_point--, cur_gene=cur_gene->Next());

  // assign our new last, find the new tail gene sequence
  last_gene = cur_gene;
  tail_gene = cur_gene->Disconnect();

  // Assign values to the new chromosone
  new_chrom->Assign(new_id,tail_gene);

  // Mark Fitness value as invalid
  fitness = -1.0;

  // create new uniq mask for each...

  return(new_chrom);
}

// Splice another Chromosone to the tail of this one
void Chromosone::Splice(Chromosone *new_tail)
{
  if (new_tail == (Chromosone *) NULL)
    {
      // hey, we gotta null tail
      return;
    }

  genelen += new_tail->genelen;   // combine their length values

  // connect the gene sequences together, change last_gene marker
  last_gene->Connect(new_tail->first_gene);
  last_gene = new_tail->last_gene;

  // Mark Fitness values as invalid
  fitness = -1.0;

  // refigure uniq masks
}

// Show gives a description of the Chromosone
void Chromosone::Show(int debug_level)
{
  if (debug_level >= 0)
    {
      cout << "[c" << chrom_id << ", f" << fitness << ", m" << mark;
      cout << ", g" << genelen << ", u" << uniqlen << "]\n";
    }
  if (debug_level > 10)
    {
      Gene *cur_gene;
      for(cur_gene = first_gene;
	  (cur_gene != (Gene *) NULL);
	  cur_gene = cur_gene->Next())
	{
	  cout << "**";
	  cur_gene->Show((debug_level - 10));
	}
      cout << endl;
    }
}

// Show gives a description of the Chromosone
// this is an overload for the debug_level = 0 version
void Chromosone::Show(void)
{
  Show(0);
}

