/*
 * Gene Definition File
 * Gene     is an instance of a Gene
 * PGene    is a pointer to an instance of a Gene (typedef'd)
 *
 * John Dunkelberg
 * created: 19 Feb 1994
 */

extern int DEBUG;

class Gene
{
  int gene_id;              // which gene am I? (mesh element)
  int allele;               // single allele (MasPar PE node)
  Gene *next_gene;          // next gene in sequence

public:
// Construct and Destruct Functions
  Gene(void);              // construct a gene
  ~Gene(void);             // destroy the gene
  int Kill(void);          // recursively kill all genes in the list
  void Assign(int, int); // assign a gene_id and allele value

// Meta Functions
  int Connect(Gene *);     // add another gene onto our tail
  Gene *Disconnect();      // terminate our tail, pass tail back
  Gene *Next();            // pass back location of our tail

// Private Info Functions
  int Gene_id(void)
    { return gene_id; }    // returns gene_id value
  int Allele(void)
    { return allele; }     // returns allele value

// Internal Functions
  Gene *Copy();            // create a copy of this gene
  void Show(int);          // provide information on the Gene
  void Show(void);         // overload of above, with 0 debug_level
};

typedef Gene *PGene;

Gene::Gene(void)
{
  gene_id = 0;
  allele = 0;
  next_gene = (Gene *) NULL;
}

Gene::~Gene(void)
{
  next_gene = (Gene *) NULL;
}

/* Recursively sends the Kill down the chain, deleting all genes */
int Gene::Kill(void)
{
  if (next_gene == (Gene *) NULL)
    return(0);

  next_gene->Kill();
  delete(next_gene);
  next_gene = (Gene *) NULL;
  return(1);
}

/* Assign actual values to the gene */
void Gene::Assign(int id, int val)
{
  gene_id = id;
  allele = val;
}

/* Add another Gene (or Gene list) on behind the current gene */
int Gene::Connect(Gene *newgene)
{
  if (newgene == (Gene *) NULL)
    return(0);

  next_gene = newgene;
  return(1);
}

// break off the tail of the gene chain here, pass back the tail pointer
Gene *Gene::Disconnect()
{
  Gene *tail;

  tail = next_gene;
  next_gene = (Gene *) NULL;
  return(tail);
}

// Pass back the location of our tail, without breaking it off
Gene *Gene::Next()
{
  return(next_gene);
}

// Create a clone copy of this Gene
Gene *Gene::Copy(void)
{
  Gene *new_gene;
  new_gene = new Gene;

  new_gene->next_gene = (Gene *) NULL;
  new_gene->gene_id = gene_id;
  new_gene->allele = allele;

  return(new_gene);
}

// describe the Gene to cout
// with the debug_level indicating depth of description
void Gene::Show(int debug_level)
{
  if (debug_level >= 0)
    {
      cout << "[" << gene_id << "," << allele << "]";

      if (debug_level >= 10)
	{
	  cout << ":" << next_gene;
	  if (next_gene != (Gene *) NULL)
	    cout << ":" << next_gene->gene_id;
	  else
	    cout << ":NULL";
	}
      cout << "\n";
    }
}

// Show(void), assumed debug level of 0
void Gene::Show(void)
{
  Show(0);
}
